@@ -1,5 +1,5 @@
 /*
-    $Id: strscan.c 40927 2013-05-25 11:37:35Z naruse $
+    $Id: strscan.c 50511 2015-05-16 12:56:48Z nobu $
 
     Copyright (c) 1999-2006 Minero Aoki
 
@@ -190,20 +190,20 @@ strscan_memsize(const void *ptr)
 
 static const rb_data_type_t strscanner_type = {
     "StringScanner",
-    {strscan_mark, strscan_free, strscan_memsize}
+    {strscan_mark, strscan_free, strscan_memsize},
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY
 };
 
 static VALUE
 strscan_s_allocate(VALUE klass)
 {
     struct strscanner *p;
+    VALUE obj = TypedData_Make_Struct(klass, struct strscanner, &strscanner_type, p);
 
-    p = ALLOC(struct strscanner);
-    MEMZERO(p, struct strscanner, 1);
     CLEAR_MATCH_STATUS(p);
     onig_region_init(&(p->regs));
     p->str = Qnil;
-    return TypedData_Wrap_Struct(klass, &strscanner_type, p);
+    return obj;
 }
 
 /*
@@ -251,7 +251,9 @@ strscan_init_copy(VALUE vself, VALUE vorig)
 	self->str = orig->str;
 	self->prev = orig->prev;
 	self->curr = orig->curr;
-	onig_region_copy(&self->regs, &orig->regs);
+	if (rb_reg_region_copy(&self->regs, &orig->regs))
+	    rb_memerror();
+	RB_GC_GUARD(vorig);
     }
 
     return vself;
@@ -976,7 +978,7 @@ strscan_matched_size(VALUE self)
 }
 
 static int
-name_to_backref_number(struct re_registers *regs, VALUE regexp, const char* name, const char* name_end)
+name_to_backref_number(struct re_registers *regs, VALUE regexp, const char* name, const char* name_end, rb_encoding *enc)
 {
     int num;
 
@@ -986,9 +988,8 @@ name_to_backref_number(struct re_registers *regs, VALUE regexp, const char* name
 	return num;
     }
     else {
-	VALUE s = rb_str_new(name, (long )(name_end - name));
-	rb_raise(rb_eIndexError, "undefined group name reference: %s",
-				 StringValuePtr(s));
+	rb_enc_raise(enc, rb_eIndexError, "undefined group name reference: %.*s",
+					  rb_long2int(name_end - name), name);
     }
 
     UNREACHABLE;
@@ -1032,13 +1033,11 @@ strscan_aref(VALUE self, VALUE idx)
 
     switch (TYPE(idx)) {
         case T_SYMBOL:
-            name = rb_id2name(SYM2ID(idx));
-            goto name_to_backref;
-            break;
+            idx = rb_sym2str(idx);
+            /* fall through */
         case T_STRING:
-            name = StringValuePtr(idx);
-        name_to_backref:
-	    i = name_to_backref_number(&(p->regs), p->regex, name, name + strlen(name));
+            RSTRING_GETMEM(idx, name, i);
+            i = name_to_backref_number(&(p->regs), p->regex, name, name + i, rb_enc_get(idx));
             break;
         default:
             i = NUM2LONG(idx);
@@ -1137,7 +1136,6 @@ strscan_restsize(VALUE self)
 }
 
 #define INSPECT_LENGTH 5
-#define BUFSIZE 256
 
 /*
  * Returns a string that represents the StringScanner object, showing:
@@ -1154,76 +1152,69 @@ static VALUE
 strscan_inspect(VALUE self)
 {
     struct strscanner *p;
-    char buf[BUFSIZE];
-    long len;
     VALUE a, b;
 
     p = check_strscan(self);
     if (NIL_P(p->str)) {
-        len = snprintf(buf, BUFSIZE, "#<%s (uninitialized)>",
-                       rb_class2name(CLASS_OF(self)));
-        return infect(rb_str_new(buf, len), p);
+	a = rb_sprintf("#<%"PRIsVALUE" (uninitialized)>", rb_obj_class(self));
+	return infect(a, p);
     }
     if (EOS_P(p)) {
-        len = snprintf(buf, BUFSIZE, "#<%s fin>",
-                       rb_class2name(CLASS_OF(self)));
-        return infect(rb_str_new(buf, len), p);
+	a = rb_sprintf("#<%"PRIsVALUE" fin>", rb_obj_class(self));
+	return infect(a, p);
     }
     if (p->curr == 0) {
-        b = inspect2(p);
-        len = snprintf(buf, BUFSIZE, "#<%s %ld/%ld @ %s>",
-                       rb_class2name(CLASS_OF(self)),
-                       p->curr, S_LEN(p),
-                       RSTRING_PTR(b));
-        return infect(rb_str_new(buf, len), p);
+	b = inspect2(p);
+	a = rb_sprintf("#<%"PRIsVALUE" %ld/%ld @ %"PRIsVALUE">",
+		       rb_obj_class(self),
+		       p->curr, S_LEN(p),
+		       b);
+	return infect(a, p);
     }
     a = inspect1(p);
     b = inspect2(p);
-    len = snprintf(buf, BUFSIZE, "#<%s %ld/%ld %s @ %s>",
-                   rb_class2name(CLASS_OF(self)),
-                   p->curr, S_LEN(p),
-                   RSTRING_PTR(a),
-                   RSTRING_PTR(b));
-    return infect(rb_str_new(buf, len), p);
+    a = rb_sprintf("#<%"PRIsVALUE" %ld/%ld %"PRIsVALUE" @ %"PRIsVALUE">",
+		   rb_obj_class(self),
+		   p->curr, S_LEN(p),
+		   a, b);
+    return infect(a, p);
 }
 
 static VALUE
 inspect1(struct strscanner *p)
 {
-    char buf[BUFSIZE];
-    char *bp = buf;
+    VALUE str;
     long len;
 
     if (p->curr == 0) return rb_str_new2("");
     if (p->curr > INSPECT_LENGTH) {
-        strcpy(bp, "..."); bp += 3;
-        len = INSPECT_LENGTH;
+	str = rb_str_new_cstr("...");
+	len = INSPECT_LENGTH;
     }
     else {
-        len = p->curr;
+	str = rb_str_new(0, 0);
+	len = p->curr;
     }
-    memcpy(bp, CURPTR(p) - len, len); bp += len;
-    return rb_str_dump(rb_str_new(buf, bp - buf));
+    rb_str_cat(str, CURPTR(p) - len, len);
+    return rb_str_dump(str);
 }
 
 static VALUE
 inspect2(struct strscanner *p)
 {
-    char buf[BUFSIZE];
-    char *bp = buf;
+    VALUE str;
     long len;
 
     if (EOS_P(p)) return rb_str_new2("");
     len = S_LEN(p) - p->curr;
     if (len > INSPECT_LENGTH) {
-        len = INSPECT_LENGTH;
-        memcpy(bp, CURPTR(p), len); bp += len;
-        strcpy(bp, "..."); bp += 3;
+	str = rb_str_new(CURPTR(p), INSPECT_LENGTH);
+	rb_str_cat2(str, "...");
     }
     else {
-        memcpy(bp, CURPTR(p), len); bp += len;
+	str = rb_str_new(CURPTR(p), len);
     }
-    return rb_str_dump(rb_str_new(buf, bp - buf));
+    return rb_str_dump(str);
 }
 
 /* =======================================================================
@@ -1334,7 +1325,7 @@ inspect2(struct strscanner *p)
  * There are aliases to several of the methods.
  */
 void
-Init_strscan()
+Init_strscan(void)
 {
     ID id_scanerr = rb_intern("ScanError");
     VALUE tmp;
@@ -1349,7 +1340,7 @@ Init_strscan()
     tmp = rb_str_new2(STRSCAN_VERSION);
     rb_obj_freeze(tmp);
     rb_const_set(StringScanner, rb_intern("Version"), tmp);
-    tmp = rb_str_new2("$Id: strscan.c 40927 2013-05-25 11:37:35Z naruse $");
+    tmp = rb_str_new2("$Id: strscan.c 50511 2015-05-16 12:56:48Z nobu $");
     rb_obj_freeze(tmp);
     rb_const_set(StringScanner, rb_intern("Id"), tmp);
 
