@@ -90,7 +90,6 @@ class OpenStruct
       hash.each_pair do |k, v|
         k = k.to_sym
         @table[k] = v
-        new_ostruct_member(k)
       end
     end
   end
@@ -99,7 +98,6 @@ class OpenStruct
   def initialize_copy(orig)
     super
     @table = @table.dup
-    @table.each_key{|key| new_ostruct_member(key)}
   end
 
   #
@@ -125,7 +123,7 @@ class OpenStruct
   #   data.each_pair.to_a  # => [[:country, "Australia"], [:population, 20000000]]
   #
   def each_pair
-    return to_enum __method__ unless block_given?
+    return to_enum(__method__) { @table.size } unless block_given?
     @table.each_pair{|p| yield p}
   end
 
@@ -141,7 +139,6 @@ class OpenStruct
   #
   def marshal_load(x)
     @table = x
-    @table.each_key{|key| new_ostruct_member(key)}
   end
 
   #
@@ -152,7 +149,7 @@ class OpenStruct
     begin
       @modifiable = true
     rescue
-      raise TypeError, "can't modify frozen #{self.class}", caller(3)
+      raise RuntimeError, "can't modify frozen #{self.class}", caller(3)
     end
     @table
   end
@@ -173,18 +170,27 @@ class OpenStruct
   end
   protected :new_ostruct_member
 
+  def respond_to_missing?(mid, include_private = false)
+    mname = mid.to_s.chomp("=").to_sym
+    @table.key?(mname) || super
+  end
+
   def method_missing(mid, *args) # :nodoc:
-    mname = mid.id2name
     len = args.length
-    if mname.chomp!('=')
+    if mname = mid[/.*(?==\z)/m]
       if len != 1
         raise ArgumentError, "wrong number of arguments (#{len} for 1)", caller(1)
       end
       modifiable[new_ostruct_member(mname)] = args[0]
     elsif len == 0
-      @table[mid]
+      if @table.key?(mid)
+        new_ostruct_member(mid)
+        @table[mid]
+      end
     else
-      raise NoMethodError, "undefined method `#{mid}' for #{self}", caller(1)
+      err = NoMethodError.new "undefined method `#{mid}' for #{self}", mid, args
+      err.set_backtrace caller(1)
+      raise err
     end
   end
 
@@ -220,8 +226,13 @@ class OpenStruct
   #
   def delete_field(name)
     sym = name.to_sym
-    singleton_class.__send__(:remove_method, sym, "#{name}=")
-    @table.delete sym
+    begin
+      singleton_class.__send__(:remove_method, sym, "#{sym}=")
+    rescue NameError
+    end
+    @table.delete(sym) do
+      raise NameError.new("no field `#{sym}' in #{self}", sym)
+    end
   end
 
   InspectKey = :__inspect_key__ # :nodoc:
