@@ -4,7 +4,6 @@ begin
   require 'pty'
 rescue LoadError
 end
-require_relative '../../ruby/envutil'
 
 class TestIO_Console < Test::Unit::TestCase
   Bug6116 = '[ruby-dev:45309]'
@@ -69,6 +68,14 @@ class TestIO_Console < Test::Unit::TestCase
     }
   end
 
+  def test_raw!
+    helper {|m, s|
+      s.raw!
+      s.print "foo\n"
+      assert_equal("foo\n", m.gets)
+    }
+  end
+
   def test_cooked
     helper {|m, s|
       assert_send([s, :echo?])
@@ -219,21 +226,30 @@ class TestIO_Console < Test::Unit::TestCase
   end
 
   if IO.console
+    def test_close
+      IO.console.close
+      assert_kind_of(IO, IO.console)
+      assert_nothing_raised(IOError) {IO.console.fileno}
+
+      IO.console(:close)
+      assert(IO.console(:tty?))
+    ensure
+      IO.console(:close)
+    end
+
     def test_sync
       assert(IO.console.sync, "console should be unbuffered")
+    ensure
+      IO.console(:close)
     end
   else
+    def test_close
+      assert_equal(["true"], run_pty("IO.console.close; p IO.console.fileno >= 0"))
+      assert_equal(["true"], run_pty("IO.console(:close); p IO.console(:tty?)"))
+    end
+
     def test_sync
-      r, w, pid = PTY.spawn(EnvUtil.rubybin, "-rio/console", "-e", "p IO.console.class")
-    rescue RuntimeError
-      skip $!
-    else
-      con = r.gets.chomp
-      Process.wait(pid)
-      assert_match("File", con)
-    ensure
-      r.close if r
-      w.close if w
+      assert_equal(["true"], run_pty("p IO.console.sync"))
     end
   end
 
@@ -248,6 +264,24 @@ class TestIO_Console < Test::Unit::TestCase
     m.close if m
     s.close if s
   end
+
+  def run_pty(src, n = 1)
+    r, w, pid = PTY.spawn(EnvUtil.rubybin, "-rio/console", "-e", src)
+  rescue RuntimeError
+    skip $!
+  else
+    result = []
+    n.times {result << r.gets.chomp}
+    Process.wait(pid)
+    if block_given?
+      yield result
+    else
+      result
+    end
+  ensure
+    r.close if r
+    w.close if w
+  end
 end if defined?(PTY) and defined?(IO::console)
 
 class TestIO_Console < Test::Unit::TestCase
@@ -263,9 +297,9 @@ class TestIO_Console < Test::Unit::TestCase
     require 'tempfile'
     NOCTTY = noctty
     def test_noctty
-      t = Tempfile.new("console")
+      t = Tempfile.new("noctty_out")
       t.close
-      t2 = Tempfile.new("console")
+      t2 = Tempfile.new("noctty_run")
       t2.close
       cmd = [*NOCTTY[1..-1],
         '-e', 'open(ARGV[0], "w") {|f|',
